@@ -7,7 +7,7 @@ from google.cloud.storage.constants import _DEFAULT_TIMEOUT
 from google_crc32c import Checksum
 
 from octue.cloud.credentials import GCPCredentialsManager
-from octue.cloud.storage.path import split, split_bucket_name_from_gs_path
+from octue.cloud.storage.path import split_bucket_name_from_gs_path
 from octue.utils.decoders import OctueJSONDecoder
 from octue.utils.encoders import OctueJSONEncoder
 
@@ -222,21 +222,29 @@ class GoogleCloudStorageClient:
         :param float timeout: time in seconds to allow for the request to complete
         :yield google.cloud.storage.blob.Blob:
         """
+        if filter is None:
+            filter = lambda blob: True
+
         if cloud_path:
             bucket_name, directory_path = split_bucket_name_from_gs_path(cloud_path)
 
         bucket = self.client.get_bucket(bucket_or_name=bucket_name)
-        blobs = bucket.list_blobs(timeout=timeout)
-        directory_path = self._strip_leading_slash(directory_path)
 
-        if filter:
-            return (
-                blob
-                for blob in blobs
-                if self._is_in_directory(blob, directory_path, include_subdirectories) and filter(blob)
-            )
+        if not directory_path.endswith("/"):
+            directory_path += "/"
 
-        return (blob for blob in blobs if self._is_in_directory(blob, directory_path, include_subdirectories))
+        if include_subdirectories:
+            blobs = bucket.list_blobs(prefix=directory_path, timeout=timeout)
+        else:
+            blobs = bucket.list_blobs(prefix=directory_path, delimiter="/", timeout=timeout)
+
+        for blob in blobs:
+            if include_subdirectories:
+                if filter(blob):
+                    yield blob
+            else:
+                if filter(blob) and not blob.name.endswith("/"):
+                    yield blob
 
     def _strip_leading_slash(self, path):
         """Strip the leading slash from a path.
@@ -246,21 +254,6 @@ class GoogleCloudStorageClient:
         """
         return path.lstrip("/")
 
-    def _is_in_directory(self, blob, directory_path, include_subdirectories=True):
-        """Check if the given blob exists in the given directory.
-
-        :param google.cloud.storage.blob.Blob blob:
-        :param str directory_path:
-        :param bool include_subdirectories: if False, subdirectories are classed as not being in the directory
-        :return bool:
-        """
-        head = split(blob.name)[0]
-
-        if include_subdirectories:
-            return directory_path in head
-
-        return head == directory_path
-
     def _blob(self, cloud_path=None, bucket_name=None, path_in_bucket=None):
         """Instantiate a blob for the given bucket at the given path. Note that this is not synced up with Google Cloud.
         Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must be provided.
