@@ -3,6 +3,7 @@ import functools
 import logging
 import os
 import tempfile
+from urllib.parse import urlparse
 import google.api_core.exceptions
 import pkg_resources
 from google_crc32c import Checksum
@@ -19,7 +20,6 @@ from octue.utils import isfile
 module_logger = logging.getLogger(__name__)
 
 
-TEMPORARY_LOCAL_FILE_CACHE = {}
 OCTUE_METADATA_NAMESPACE = "octue"
 
 ID_DEFAULT = None
@@ -46,10 +46,13 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
           "sha-512/256": "somesha"
         },
 
+    :param str|None path: The path of this file locally or in the cloud, which may include folders or subfolders, within the dataset. If no path_from parameter is set, then absolute paths are acceptable, otherwise relative paths are required.
+    :param str|None local_path: If a cloud path is given as the `path` parameter, this is the path to an existing local file that is known to be in sync with the cloud object
+    :param str|None cloud_path: If a local path is given for the `path` parameter, this is a cloud path to keep in sync with the local file
+    :param str|None project_name: The name of the project if the datafile also exists in the cloud
     :param datetime.datetime|int|float|None timestamp: A posix timestamp associated with the file, in seconds since epoch, typically when it was created but could relate to a relevant time point for the data
     :param str id: The Universally Unique ID of this file (checked to be valid if not None, generated if None)
     :param logging.Logger logger: A logger instance to which operations with this datafile will be logged. Defaults to the module logger.
-    :param Union[str, path-like] path: The path of this file, which may include folders or subfolders, within the dataset. If no path_from parameter is set, then absolute paths are acceptable, otherwise relative paths are required.
     :param Pathable path_from: The root Pathable object (typically a Dataset) that this Datafile's path is relative to.
     :param str|None project_name: The name of the cloud project if the datafile is located in the cloud
     :param dict|TagDict tags: key-value pairs with string keys conforming to the Octue tag format (see TagDict)
@@ -65,6 +68,8 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
         "id",
         "name",
         "path",
+        "cloud_path",
+        "project_name",
         "tags",
         "labels",
         "timestamp",
@@ -74,11 +79,13 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
     def __init__(
         self,
         path,
+        local_path=None,
+        cloud_path=None,
+        project_name=None,
         timestamp=None,
         id=ID_DEFAULT,
         logger=None,
         path_from=None,
-        project_name=None,
         tags=TAGS_DEFAULT,
         labels=LABELS_DEFAULT,
         skip_checks=True,
@@ -99,28 +106,53 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
         )
 
         self.timestamp = timestamp
-        self.extension = self._get_extension_from_path()
+        self.extension = os.path.splitext(path)[-1].strip(".")
+        self.project_name = project_name
+
+        self._local_path = None
+        self._cloud_path = None
         self._hypothetical = hypothetical
         self._open_attributes = {"mode": mode, "update_cloud_metadata": update_cloud_metadata, **kwargs}
-        self._cloud_metadata = {"project_name": project_name}
+        self._cloud_metadata = {}
 
-        if self.is_in_cloud and not self._hypothetical:
+        if self.path.startswith(CLOUD_STORAGE_PROTOCOL):
+            if project_name is None:
+                raise CloudLocationNotSpecified(
+                    f"The `project_name` parameter is required to instantiate a Datafile from a cloud object; received "
+                    f"{project_name}."
+                )
 
-            # Collect any non-`None` metadata instantiation parameters so the user can be warned if they conflict with
-            # any metadata already on the cloud object.
-            initialisation_parameters = {}
+            self._store_cloud_location(project_name=project_name, cloud_path=path)
 
-            for parameter in ("id", "timestamp", "tags", "labels"):
-                value = locals().get(parameter)
-                if value is not None:
-                    initialisation_parameters[parameter] = value
+            if not self._hypothetical:
+                # Collect any non-`None` metadata instantiation parameters so the user can be warned if they conflict
+                # with any metadata already on the cloud object.
+                initialisation_parameters = {}
 
-            self._use_cloud_metadata(**initialisation_parameters)
-            return
+                for parameter in ("id", "timestamp", "tags", "labels"):
+                    value = locals().get(parameter)
+                    if value is not None:
+                        initialisation_parameters[parameter] = value
 
-        # Run integrity checks on the file
-        if not skip_checks:
-            self.check(**kwargs)
+                self._use_cloud_metadata(**initialisation_parameters)
+
+            if local_path:
+                # If there is no file at the given local path or the file is different to the one in the cloud, download
+                # the cloud file locally.
+                if not os.path.exists(local_path) or self._cloud_metadata.get("crc32c") != calculate_hash(local_path):
+                    self.download(local_path)
+                else:
+                    self._local_path = local_path
+
+        else:
+            self._local_path = self.absolute_path
+
+            # Run integrity checks on the file.
+            if not skip_checks:
+                self.check(**kwargs)
+
+            if cloud_path:
+                self.cloud_path = cloud_path
 
     def __enter__(self):
         self._open_context_manager = self.open(**self._open_attributes)
@@ -166,8 +198,8 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
     def to_cloud(
         self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None, update_cloud_metadata=True
     ):
-        """Upload a datafile to Google Cloud Storage. Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must be
-        provided.
+        """Upload a datafile to Google Cloud Storage. Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must
+        be provided.
 
         :param str|None project_name: name of Google Cloud project to store datafile in
         :param str|None cloud_path: full path to cloud storage location to store datafile at (e.g. `gs://bucket_name/path/to/file.csv`)
@@ -176,18 +208,14 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
         :param bool update_cloud_metadata: if `True`, update the metadata of the datafile in the cloud at upload time
         :return str: gs:// path for datafile
         """
-        project_name, bucket_name, path_in_bucket = self._get_cloud_location(
-            project_name, cloud_path, bucket_name, path_in_bucket
-        )
-
-        self.get_cloud_metadata(project_name, bucket_name=bucket_name, path_in_bucket=path_in_bucket)
+        project_name, cloud_path = self._get_cloud_location(project_name, cloud_path, bucket_name, path_in_bucket)
+        self.get_cloud_metadata()
 
         # If the datafile's file has been changed locally, overwrite its cloud copy.
         if self._cloud_metadata.get("crc32c") != self.hash_value:
             GoogleCloudStorageClient(project_name=project_name).upload_file(
-                local_path=self.get_local_path(),
-                bucket_name=bucket_name,
-                path_in_bucket=path_in_bucket,
+                local_path=self.local_path,
+                cloud_path=cloud_path,
                 metadata=self.metadata(),
             )
 
@@ -196,57 +224,97 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
             local_metadata = self.metadata()
 
             if self._cloud_metadata.get("custom_metadata") != local_metadata:
-                self.update_cloud_metadata(project_name, bucket_name=bucket_name, path_in_bucket=path_in_bucket)
+                self.update_cloud_metadata()
 
-        return cloud_path or storage.path.generate_gs_path(bucket_name, path_in_bucket)
+        return self.cloud_path
 
-    def get_cloud_metadata(self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None):
-        """Get the cloud metadata for the datafile. Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must be
-        provided.
+    def get_cloud_metadata(self):
+        """Get the cloud metadata for the datafile.
 
-        :param str|None project_name:
-        :param str|None cloud_path:
-        :param str|None bucket_name:
-        :param str|None path_in_bucket:
         :return dict:
         """
-        project_name, bucket_name, path_in_bucket = self._get_cloud_location(
-            project_name, cloud_path, bucket_name, path_in_bucket
-        )
+        if not self.cloud_path:
+            self._raise_cloud_location_error()
 
-        cloud_metadata = GoogleCloudStorageClient(project_name).get_metadata(
-            bucket_name=bucket_name, path_in_bucket=path_in_bucket
-        )
+        cloud_metadata = GoogleCloudStorageClient(self.project_name).get_metadata(cloud_path=self.cloud_path)
 
-        if cloud_metadata is None:
-            return None
+        if cloud_metadata:
+            self._cloud_metadata = cloud_metadata
 
-        self._cloud_metadata = cloud_metadata
+    def update_cloud_metadata(self):
+        """Update the cloud metadata for the datafile.
 
-    def update_cloud_metadata(self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None):
-        """Update the cloud metadata for the datafile. Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must be
-        provided.
-
-        :param str|None project_name:
-        :param str|None cloud_path:
-        :param str|None bucket_name:
-        :param str|None path_in_bucket:
         :return None:
         """
-        project_name, bucket_name, path_in_bucket = self._get_cloud_location(
-            project_name, cloud_path, bucket_name, path_in_bucket
-        )
+        if not self.cloud_path:
+            self._raise_cloud_location_error()
 
-        GoogleCloudStorageClient(project_name=project_name).overwrite_custom_metadata(
-            metadata=self.metadata(),
-            bucket_name=bucket_name,
-            path_in_bucket=path_in_bucket,
+        GoogleCloudStorageClient(project_name=self.project_name).overwrite_custom_metadata(
+            metadata=self.metadata(), cloud_path=self.cloud_path
         )
 
     @property
     def name(self):
         return self._name or str(os.path.split(self.path)[-1])
 
+    @property
+    def cloud_path(self):
+        """Get the cloud path of the datafile.
+
+        :return str|None:
+        """
+        return self._cloud_path
+
+    @cloud_path.setter
+    def cloud_path(self, path):
+        """Set the cloud path of the datafile.
+
+        :param str|None path:
+        :return None:
+        """
+        if path is None:
+
+            if not self.exists_locally:
+                raise CloudLocationNotSpecified(
+                    "The cloud path cannot be reset because this datafile only exists in the cloud."
+                )
+
+            self._cloud_path = None
+
+        else:
+            self.to_cloud(cloud_path=path)
+
+    @property
+    def cloud_protocol(self):
+        """Get the cloud protocol of the datafile if it exists in the cloud (e.g. "gs" for a cloud path of
+        "gs://my-bucket/my-file.txt").
+
+        :return str|None:
+        """
+        if not self.exists_in_cloud:
+            return None
+        return urlparse(self.cloud_path).scheme
+
+    @property
+    def bucket_name(self):
+        """Get the name of the bucket the datafile exists in if it exists in the cloud.
+
+        :return str|None:
+        """
+        if self.cloud_path:
+            return storage.path.split_bucket_name_from_gs_path(self.cloud_path)[0]
+        return None
+
+    @property
+    def path_in_bucket(self):
+        """Get the path of the datafile in its bucket if it exists in the cloud.
+
+        :return str|None:
+        """
+        if self.cloud_path:
+            return storage.path.split_bucket_name_from_gs_path(self.cloud_path)[1]
+        return None
+
     @property
     def timestamp(self):
         return self._timestamp
@@ -303,30 +371,45 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
         return os.path.getsize(self.absolute_path)
 
     @property
-    def is_in_cloud(self):
+    def exists_in_cloud(self):
         """Does the file exist in the cloud?
 
         :return bool:
         """
-        return self.path.startswith(CLOUD_STORAGE_PROTOCOL)
+        return self.cloud_path is not None
 
-    def get_local_path(self):
-        """Get the local path for the datafile, downloading it from the cloud to a temporary file if necessary. If
-        downloaded, the local path is added to a cache to avoid downloading again in the same runtime.
+    @property
+    def exists_locally(self):
+        """Does the file exist locally?
 
-        :return str:
+        :return bool:
         """
-        if not self.is_in_cloud:
-            return self.absolute_path
+        return self._local_path is not None
 
-        if self.absolute_path in TEMPORARY_LOCAL_FILE_CACHE:
-            return TEMPORARY_LOCAL_FILE_CACHE[self.absolute_path]
+    def download(self, local_path=None):
+        """Download the file from the cloud to the given local path or a random temporary path.
 
-        temporary_local_path = tempfile.NamedTemporaryFile(delete=False).name
+        :param str|None local_path:
+        :raise CloudLocationNotSpecified: if the datafile does not exist in the cloud
+        :return str: path to local file
+        """
+        if not self.exists_in_cloud:
+            raise CloudLocationNotSpecified("Cannot download a file that doesn't exist in the cloud.")
+
+        # Avoid downloading to a local path if the datafile has already been downloaded to it.
+        if (local_path is None and self._local_path is not None) or (
+            local_path is not None and local_path == self._local_path
+        ):
+            return self._local_path
+
+        if local_path is not None:
+            self._local_path = os.path.abspath(local_path)
+        else:
+            self._local_path = tempfile.NamedTemporaryFile(delete=False).name
 
         try:
-            GoogleCloudStorageClient(project_name=self._cloud_metadata["project_name"]).download_to_file(
-                local_path=temporary_local_path, cloud_path=self.absolute_path
+            GoogleCloudStorageClient(project_name=self.project_name).download_to_file(
+                local_path=self._local_path, cloud_path=self.cloud_path
             )
 
         except google.api_core.exceptions.NotFound as e:
@@ -334,32 +417,62 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
             if self._open_attributes["mode"] == "r":
                 raise e
 
-        TEMPORARY_LOCAL_FILE_CACHE[self.absolute_path] = temporary_local_path
-
         # Now use hash value of local file instead of cloud file.
         self.reset_hash()
-        return temporary_local_path
+        return self._local_path
 
-    def clear_from_file_cache(self):
-        """Clear the datafile from the temporary local file cache, if it is in there. If datafile.get_local_path is
-        called again and the datafile is a cloud datafile, the file will be re-downloaded to a new temporary local path,
-        allowing any independent cloud updates to be synced locally.
+    @property
+    def local_path(self):
+        """Get the local path for the datafile, downloading it from the cloud to a temporary file if necessary. If
+        downloaded, the local path is added to a cache to avoid downloading again in the same runtime.
+        """
+        if self._local_path:
+            return self._local_path
 
+        return self.download()
+
+    @local_path.setter
+    def local_path(self, path):
+        """Set the local path of the datafile to an empty path (a path not corresponding to an existing file) and
+        download the contents of the corresponding cloud file to the local path.
+
+        :param str path:
+        :raise FileExistsError: if the path corresponds to an existing file
         :return None:
         """
-        if self.absolute_path in TEMPORARY_LOCAL_FILE_CACHE:
-            del TEMPORARY_LOCAL_FILE_CACHE[self.absolute_path]
+        if path is None:
+            if not self.exists_in_cloud:
+                raise CloudLocationNotSpecified(
+                    "The local path cannot be reset because this datafile only exists locally."
+                )
+
+            self._local_path = None
+            return
+
+        if os.path.exists(path):
+            raise FileExistsError(
+                "Only a path not corresponding to an existing file can be used. This is because the contents of the "
+                "existing cloud file will be downloaded to the new local path and would overwrite any existing file at "
+                "the given path."
+            )
+
+        GoogleCloudStorageClient(project_name=self.project_name).download_to_file(
+            local_path=path, cloud_path=self.cloud_path
+        )
+
+        self._local_path = os.path.abspath(path)
 
     def _use_cloud_metadata(self, **initialisation_parameters):
-        """Populate the datafile's attributes from the cloud location defined by its path (by necessity a cloud path)
-        and project name.
+        """Populate the datafile's attributes from the metadata of the cloud object located at its path (by necessity a
+        cloud path) and project name. If there is a conflict between the cloud metadata and a given local initialisation
+        parameter, the local value is used.
 
         :param initialisation_parameters: key-value pairs of initialisation parameter names and values (provide to check for conflicts with cloud metadata)
         :return None:
         """
-        self.get_cloud_metadata(project_name=self._cloud_metadata["project_name"], cloud_path=self.path)
+        self.get_cloud_metadata()
         cloud_custom_metadata = self._cloud_metadata.get("custom_metadata", {})
-        self._check_for_attribute_conflict(cloud_custom_metadata, **initialisation_parameters)
+        self._warn_about_attribute_conflicts(cloud_custom_metadata, **initialisation_parameters)
 
         self._set_id(
             initialisation_parameters.get(
@@ -378,7 +491,7 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
                 ),
             )
 
-    def _check_for_attribute_conflict(self, cloud_custom_metadata, **initialisation_parameters):
+    def _warn_about_attribute_conflicts(self, cloud_custom_metadata, **initialisation_parameters):
         """Raise a warning if there is a conflict between the cloud custom metadata and the given initialisation
         parameters if the cloud value is not `None` or an empty collection.
 
@@ -398,20 +511,9 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
                 f"value has been used and will overwrite the cloud value if the datafile is saved."
             )
 
-    def _get_extension_from_path(self, path=None):
-        """Gets extension of a file, either from a provided file path or from self.path field"""
-        path = path or self.path
-        return os.path.splitext(path)[-1].strip(".")
-
     def _calculate_hash(self):
         """Calculate the hash of the file."""
-        hash = Checksum()
-
-        with open(self.get_local_path(), "rb") as f:
-            # Read and update hash value in blocks of 4K.
-            for byte_block in iter(lambda: f.read(4096), b""):
-                hash.update(byte_block)
-
+        hash = calculate_hash(self.local_path)
         return super()._calculate_hash(hash)
 
     def _get_cloud_location(self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None):
@@ -423,37 +525,51 @@ class Datafile(Labelable, Taggable, Serialisable, Pathable, Loggable, Identifiab
         :param str|None cloud_path:
         :param str|None bucket_name:
         :param str|None path_in_bucket:
-        :raise octue.exceptions.CloudLocationNotSpecified: if an exact cloud location isn't provided and isn't available
-            implicitly (i.e. the Datafile wasn't loaded from the cloud previously)
-        :return (str, str, str):
+        :raise octue.exceptions.CloudLocationNotSpecified: if an exact cloud location isn't provided and isn't available implicitly (i.e. the Datafile wasn't loaded from the cloud previously)
+        :return (str, str): project name and cloud path
         """
-        if cloud_path:
-            bucket_name, path_in_bucket = storage.path.split_bucket_name_from_gs_path(cloud_path)
+        project_name = project_name or self.project_name
+        cloud_path = cloud_path or self.cloud_path
 
-        try:
-            project_name = project_name or self._cloud_metadata["project_name"]
-            bucket_name = bucket_name or self._cloud_metadata["bucket_name"]
-            path_in_bucket = path_in_bucket or self._cloud_metadata["path_in_bucket"]
-        except KeyError:
-            raise CloudLocationNotSpecified(
-                f"{self!r} wasn't previously loaded from the cloud so doesn't have an implicit cloud location - please"
-                f"specify its exact location (its project_name, bucket_name, and path_in_bucket)."
-            )
+        if not cloud_path:
+            try:
+                cloud_path = storage.path.generate_gs_path(bucket_name, path_in_bucket)
+            except (TypeError, AttributeError):
+                self._raise_cloud_location_error()
 
-        self._store_cloud_location(project_name, bucket_name, path_in_bucket)
-        return project_name, bucket_name, path_in_bucket
+        self._store_cloud_location(project_name=project_name, cloud_path=cloud_path)
+        return project_name, cloud_path
 
-    def _store_cloud_location(self, project_name, bucket_name, path_in_bucket):
-        """Store the cloud location of the datafile.
+    def _store_cloud_location(self, project_name=None, cloud_path=None, bucket_name=None, path_in_bucket=None):
+        """Store the cloud location of the datafile. Either (`bucket_name` and `path_in_bucket`) or `cloud_path` must be
+        provided.
 
-        :param str project_name:
-        :param str bucket_name:
-        :param str path_in_bucket:
+        :param str|None project_name:
+        :param str|None cloud_path:
+        :param str|None bucket_name:
+        :param str|None path_in_bucket:
         :return None:
         """
-        self._cloud_metadata["project_name"] = project_name
-        self._cloud_metadata["bucket_name"] = bucket_name
-        self._cloud_metadata["path_in_bucket"] = path_in_bucket
+        self.project_name = project_name
+
+        if cloud_path:
+            self._cloud_path = cloud_path
+        else:
+            try:
+                self._cloud_path = storage.path.generate_gs_path(bucket_name, path_in_bucket)
+            except (TypeError, AttributeError):
+                self._raise_cloud_location_error()
+
+    def _raise_cloud_location_error(self):
+        """Raise an error indicating that the cloud location of the datafile has not yet been specified.
+
+        :raise CloudLocationNotSpecified:
+        :return None:
+        """
+        raise CloudLocationNotSpecified(
+            f"{self!r} wasn't previously loaded from the cloud so doesn't have an implicit cloud location - please "
+            f"specify its exact location (its project name and cloud path)."
+        )
 
     def check(self, size_bytes=None, sha=None, last_modified=None, extension=None):
         """Check file presence and integrity"""
@@ -524,10 +640,8 @@ class _DatafileContextManager:
     ```
 
     :param octue.resources.datafile.Datafile datafile:
-    :param str mode: open the datafile for reading/editing in this mode (the mode options are the same as for the
-        builtin open function)
-    :param bool update_cloud_metadata: this is True, update the cloud metadata of
-        the datafile when the context is exited
+    :param str mode: open the datafile for reading/editing in this mode (the mode options are the same as for the builtin `open` function)
+    :param bool update_cloud_metadata: if this is True, update the cloud metadata of the datafile when the context is exited
     :return None:
     """
 
@@ -539,19 +653,16 @@ class _DatafileContextManager:
         self._update_cloud_metadata = update_cloud_metadata
         self.kwargs = kwargs
         self._fp = None
-        self.path = None
 
     def __enter__(self):
         """Open the datafile, first downloading it from the cloud if necessary.
 
         :return io.TextIOWrapper:
         """
-        self.path = self.datafile.get_local_path()
-
         if "w" in self.mode:
-            os.makedirs(os.path.split(self.path)[0], exist_ok=True)
+            os.makedirs(os.path.split(self.datafile.local_path)[0], exist_ok=True)
 
-        self._fp = open(self.path, self.mode, **self.kwargs)
+        self._fp = open(self.datafile.local_path, self.mode, **self.kwargs)
         return self._fp
 
     def __exit__(self, *args):
@@ -563,5 +674,17 @@ class _DatafileContextManager:
         if self._fp is not None:
             self._fp.close()
 
-        if self.datafile.is_in_cloud and any(character in self.mode for character in self.MODIFICATION_MODES):
+        if self.datafile.exists_in_cloud and any(character in self.mode for character in self.MODIFICATION_MODES):
             self.datafile.to_cloud(update_cloud_metadata=self._update_cloud_metadata)
+
+
+def calculate_hash(path):
+    """Calculate the hash of the file at the given path."""
+    hash = Checksum()
+
+    with open(path, "rb") as f:
+        # Read and update hash value in blocks of 4K.
+        for byte_block in iter(lambda: f.read(4096), b""):
+            hash.update(byte_block)
+
+    return hash
