@@ -36,7 +36,10 @@ def run(analysis, *args, **kwargs):
     #
     # See the Dataset class help for more.
     metadata_file = input_dataset.get_file_by_label("meta")
-    timeseries_files = input_dataset.get_file_sequence(labels__contains="timeseries")
+
+    timeseries_files = input_dataset.files.filter(labels__contains="timeseries").order_by(
+        "tags__sequence", check_start_value=0, check_constant_increment=1
+    )
     #
     # We used these because they're special helpers - in this case ensuring that there's only one metadata file and
     # ensuring that the timeseries files come in a strictly ordered sequence.
