@@ -245,7 +245,7 @@ class MockService(Service):
         # Ignore any errors from the answering service as they will be raised on the remote service in practice, not
         # locally as is done in this mock.
         if input_manifest is not None:
-            input_manifest = input_manifest.serialise(to_string=True)
+            input_manifest = input_manifest.serialise()
 
         try:
             self.children[service_id].answer(
