@@ -171,17 +171,17 @@ class TestUploadFileToGoogleCloud(BaseTestCase):
         self.assertEqual(len(contents), 1)
         self.assertEqual(contents[0].name, storage.path.join(directory_path, self.FILENAME))
 
-    def test_scandir_with_gs_path(self):
-        """Test that Google Cloud storage "directories"' contents can be listed when a GS path is used."""
-        directory_path = storage.path.join("my", "path")
-        path_in_bucket = storage.path.join(directory_path, self.FILENAME)
-        gs_path = f"gs://{TEST_BUCKET_NAME}/{path_in_bucket}"
+    def test_scandir_with_cloud_path(self):
+        """Test that Google Cloud storage "directories"' contents can be listed when a cloud path is used."""
+        cloud_directory_path = storage.path.generate_gs_path(TEST_BUCKET_NAME, "a", "path")
 
-        self.storage_client.upload_from_string(string=json.dumps({"height": 32}), cloud_path=gs_path)
-        contents = list(self.storage_client.scandir(gs_path))
+        self.storage_client.upload_from_string(
+            string=json.dumps({"height": 32}), cloud_path=storage.path.join(cloud_directory_path, self.FILENAME)
+        )
 
+        contents = list(self.storage_client.scandir(cloud_directory_path))
         self.assertEqual(len(contents), 1)
-        self.assertEqual(contents[0].name, storage.path.join(directory_path, self.FILENAME))
+        self.assertEqual(contents[0].name, "a/path/my_file.txt")
 
     def test_scandir_with_empty_directory(self):
         """Test that an empty directory shows as such."""
@@ -189,6 +189,59 @@ class TestUploadFileToGoogleCloud(BaseTestCase):
         contents = list(self.storage_client.scandir(bucket_name=TEST_BUCKET_NAME, directory_path=directory_path))
         self.assertEqual(len(contents), 0)
 
+    def test_scandir_with_directory_of_subdirectories_includes_subdirectories_by_default(self):
+        """Test that subdirectories of the given directory are included by scandir by default."""
+        directory_path = storage.path.join("the", "path")
+
+        self.storage_client.upload_from_string(
+            string=json.dumps({"height": 32}),
+            bucket_name=TEST_BUCKET_NAME,
+            path_in_bucket=storage.path.join(directory_path, self.FILENAME),
+        )
+
+        # Add a file in a subdirectory.
+        self.storage_client.upload_from_string(
+            string=json.dumps({"height": 32}),
+            bucket_name=TEST_BUCKET_NAME,
+            path_in_bucket=storage.path.join(directory_path, "sub_directory", "blah.txt"),
+        )
+
+        contents = list(self.storage_client.scandir(bucket_name=TEST_BUCKET_NAME, directory_path=directory_path))
+        self.assertEqual(len(contents), 2)
+
+        self.assertEqual(
+            {blob.name for blob in contents},
+            {
+                storage.path.join(directory_path, self.FILENAME),
+                storage.path.join(directory_path, "sub_directory", "blah.txt"),
+            },
+        )
+
+    def test_scandir_with_directory_of_subdirectories_ignores_subdirectories_if_told_to(self):
+        """Test that subdirectories of the given directory are ignored by scandir if told to."""
+        directory_path = storage.path.join("my", "path")
+
+        self.storage_client.upload_from_string(
+            string=json.dumps({"height": 32}),
+            bucket_name=TEST_BUCKET_NAME,
+            path_in_bucket=storage.path.join(directory_path, self.FILENAME),
+        )
+
+        # Add a file in a subdirectory.
+        self.storage_client.upload_from_string(
+            string=json.dumps({"height": 32}),
+            bucket_name=TEST_BUCKET_NAME,
+            path_in_bucket=storage.path.join(directory_path, "sub_directory", "blah.txt"),
+        )
+
+        contents = list(
+            self.storage_client.scandir(
+                bucket_name=TEST_BUCKET_NAME, directory_path=directory_path, include_subdirectories=False
+            )
+        )
+        self.assertEqual(len(contents), 1)
+        self.assertEqual(contents[0].name, storage.path.join(directory_path, self.FILENAME))
+
     def test_get_metadata(self):
         """Test that file metadata can be retrieved."""
         self.storage_client.upload_from_string(
