@@ -4,7 +4,7 @@ import os
 import tempfile
 import warnings
 
-from octue import exceptions
+from octue import definitions, exceptions
 from octue.cloud import storage
 from octue.cloud.storage import GoogleCloudStorageClient
 from octue.resources import Datafile, Dataset
@@ -242,7 +242,7 @@ class DatasetTestCase(BaseTestCase):
     def test_serialise(self):
         """Test that a dataset can be serialised."""
         dataset = self.create_valid_dataset()
-        self.assertEqual(len(dataset.serialise()["files"]), 2)
+        self.assertEqual(len(dataset.to_primitive()["files"]), 2)
 
     def test_exists_in_cloud(self):
         """Test whether all files of a dataset are in the cloud or not can be determined."""
@@ -313,14 +313,16 @@ class DatasetTestCase(BaseTestCase):
                     self.assertEqual(file.path, f"gs://{TEST_BUCKET_NAME}/a_directory/{dataset.name}/{file.name}")
 
     def test_from_cloud_with_no_datafile_json_file(self):
-        """Test that any cloud directory can be accessed as a dataset, even if it has no `dataset.json` file in it. Also
-        test that the cloud dataset doesn't lose any information during serialization.
+        """Test that any cloud directory can be accessed as a dataset if it has no `dataset.json` metadata file in it,
+        the cloud dataset doesn't lose any information during serialization, and a metadata file is uploaded afterwards.
         """
-        GoogleCloudStorageClient(TEST_PROJECT_NAME).upload_from_string(
+        cloud_storage_client = GoogleCloudStorageClient(TEST_PROJECT_NAME)
+
+        cloud_storage_client.upload_from_string(
             "[1, 2, 3]", bucket_name=TEST_BUCKET_NAME, path_in_bucket="my_dataset/file_0.txt"
         )
 
-        GoogleCloudStorageClient(TEST_PROJECT_NAME).upload_from_string(
+        cloud_storage_client.upload_from_string(
             "[4, 5, 6]", bucket_name=TEST_BUCKET_NAME, path_in_bucket="my_dataset/file_1.txt"
         )
 
@@ -337,12 +339,33 @@ class DatasetTestCase(BaseTestCase):
             self.assertEqual(file.path, f"gs://{TEST_BUCKET_NAME}/my_dataset/{file.name}")
 
         # Test serialisation doesn't lose any information.
-        deserialised_dataset = Dataset.deserialise(cloud_dataset.serialise())
+        deserialised_dataset = Dataset.deserialise(cloud_dataset.to_primitive())
         self.assertEqual(deserialised_dataset.id, cloud_dataset.id)
         self.assertEqual(deserialised_dataset.name, cloud_dataset.name)
         self.assertEqual(deserialised_dataset.path, cloud_dataset.path)
         self.assertEqual(deserialised_dataset.hash_value, cloud_dataset.hash_value)
 
+        # Test dataset metadata file has been uploaded.
+        dataset_metadata = json.loads(
+            cloud_storage_client.download_as_string(
+                cloud_path=storage.path.join(cloud_dataset.path, definitions.DATASET_METADATA_FILENAME)
+            )
+        )
+        del dataset_metadata["id"]
+
+        self.assertEqual(
+            dataset_metadata,
+            {
+                "files": [
+                    "gs://octue-test-bucket/my_dataset/file_0.txt",
+                    "gs://octue-test-bucket/my_dataset/file_1.txt",
+                ],
+                "labels": [],
+                "name": "my_dataset",
+                "tags": {},
+            },
+        )
+
     def test_to_cloud(self):
         """Test that a dataset can be uploaded to the cloud via (`bucket_name`, `output_directory`) and via `gs_path`,
         including all its files and a serialised JSON file of the Datafile instance.
@@ -367,19 +390,18 @@ class DatasetTestCase(BaseTestCase):
 
             bucket_name = TEST_BUCKET_NAME
             output_directory = "my_datasets"
-            gs_path = storage.path.generate_gs_path(bucket_name, output_directory)
+            cloud_path = storage.path.generate_gs_path(bucket_name, output_directory)
 
             for location_parameters in (
                 {"bucket_name": bucket_name, "output_directory": output_directory, "cloud_path": None},
-                {"bucket_name": None, "output_directory": None, "cloud_path": gs_path},
+                {"bucket_name": None, "output_directory": None, "cloud_path": cloud_path},
             ):
                 dataset.to_cloud(TEST_PROJECT_NAME, **location_parameters)
 
                 storage_client = GoogleCloudStorageClient(TEST_PROJECT_NAME)
 
                 persisted_file_0 = storage_client.download_as_string(
-                    bucket_name=TEST_BUCKET_NAME,
-                    path_in_bucket=storage.path.join(output_directory, dataset.name, "file_0.txt"),
+                    cloud_path=storage.path.join(cloud_path, dataset.name, "file_0.txt"),
                 )
 
                 self.assertEqual(persisted_file_0, "[1, 2, 3]")
@@ -393,7 +415,9 @@ class DatasetTestCase(BaseTestCase):
                 persisted_dataset = json.loads(
                     storage_client.download_as_string(
                         bucket_name=TEST_BUCKET_NAME,
-                        path_in_bucket=storage.path.join(output_directory, dataset.name, "dataset.json"),
+                        path_in_bucket=storage.path.join(
+                            output_directory, dataset.name, definitions.DATASET_METADATA_FILENAME
+                        ),
                     )
                 )
 
@@ -405,7 +429,7 @@ class DatasetTestCase(BaseTestCase):
                     ],
                 )
 
-                self.assertEqual(persisted_dataset["tags"], dataset.tags.serialise())
+                self.assertEqual(persisted_dataset["tags"], dataset.tags.to_primitive())
 
     def test_download_all_files(self):
         """Test that all files in a dataset can be downloaded with one command."""
